"use strict";
/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `PrestoDriver` and related types declaration.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PrestoDriver = void 0;
const base_driver_1 = require("@cubejs-backend/base-driver");
const shared_1 = require("@cubejs-backend/shared");
const stream_1 = require("stream");
const ramda_1 = require("ramda");
const sqlstring_1 = __importDefault(require("sqlstring"));
const presto = require('presto-client');
const SUPPORTED_BUCKET_TYPES = ['gcs', 's3'];
/**
 * Presto driver class.
 */
class PrestoDriver extends base_driver_1.BaseDriver {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 2;
    }
    config;
    catalog;
    client;
    useSelectTestConnection;
    /**
     * Class constructor.
     */
    constructor(config = {}) {
        super();
        const dataSource = config.dataSource ||
            (0, shared_1.assertDataSource)('default');
        const dbUser = (0, shared_1.getEnv)('dbUser', { dataSource });
        const dbPassword = (0, shared_1.getEnv)('dbPass', { dataSource });
        const authToken = (0, shared_1.getEnv)('prestoAuthToken', { dataSource });
        if (authToken && dbPassword) {
            throw new Error('Both user/password and auth token are set. Please remove password or token.');
        }
        this.useSelectTestConnection = (0, shared_1.getEnv)('dbUseSelectTestConnection', { dataSource });
        this.config = {
            host: (0, shared_1.getEnv)('dbHost', { dataSource }),
            port: (0, shared_1.getEnv)('dbPort', { dataSource }),
            catalog: (0, shared_1.getEnv)('prestoCatalog', { dataSource }) ||
                (0, shared_1.getEnv)('dbCatalog', { dataSource }),
            schema: (0, shared_1.getEnv)('dbName', { dataSource }) ||
                (0, shared_1.getEnv)('dbSchema', { dataSource }),
            user: dbUser,
            ...(authToken ? { custom_auth: `Bearer ${authToken}` } : {}),
            ...(dbPassword ? { basic_auth: { user: dbUser, password: dbPassword } } : {}),
            ssl: this.getSslOptions(dataSource),
            bucketType: (0, shared_1.getEnv)('dbExportBucketType', { supported: SUPPORTED_BUCKET_TYPES, dataSource }),
            exportBucket: (0, shared_1.getEnv)('dbExportBucket', { dataSource }),
            accessKeyId: (0, shared_1.getEnv)('dbExportBucketAwsKey', { dataSource }),
            secretAccessKey: (0, shared_1.getEnv)('dbExportBucketAwsSecret', { dataSource }),
            exportBucketRegion: (0, shared_1.getEnv)('dbExportBucketAwsRegion', { dataSource }),
            credentials: (0, shared_1.getEnv)('dbExportGCSCredentials', { dataSource }),
            ...config
        };
        this.catalog = this.config.catalog;
        this.client = new presto.Client(this.config);
    }
    async testConnection() {
        if (this.useSelectTestConnection) {
            return this.testConnectionViaSelect();
        }
        return new Promise((resolve, reject) => {
            // Get node list of presto cluster and return it.
            // @see https://prestodb.io/docs/current/rest/node.html
            this.client.nodes(null, (error, _result) => {
                if (error) {
                    reject(error);
                }
                else {
                    resolve();
                }
            });
        });
    }
    async testConnectionViaSelect() {
        const query = sqlstring_1.default.format('SELECT 1', []);
        await this.queryPromised(query, false);
    }
    query(query, values) {
        return this.queryPromised(this.prepareQueryWithParams(query, values), false);
    }
    prepareQueryWithParams(query, values) {
        return sqlstring_1.default.format(query, (values || []).map(value => (typeof value === 'string' ? {
            toSqlString: () => sqlstring_1.default.escape(value).replace(/\\\\([_%])/g, '\\$1'),
        } : value)));
    }
    queryPromised(query, streaming) {
        const toError = (error) => new Error(error.error ? `${error.message}\n${error.error}` : error.message);
        if (streaming) {
            const rowStream = new stream_1.Transform({
                writableObjectMode: true,
                readableObjectMode: true,
                transform(obj, encoding, callback) {
                    callback(null, obj);
                }
            });
            return new Promise((resolve, reject) => {
                this.client.execute({
                    query,
                    schema: this.config.schema || 'default',
                    session: this.config.queryTimeout ? `query_max_run_time=${this.config.queryTimeout}s` : undefined,
                    columns: (error, columns) => {
                        resolve({
                            rowStream,
                            types: columns
                        });
                    },
                    data: (error, data, columns) => {
                        const normalData = this.normalizeResultOverColumns(data, columns);
                        for (const obj of normalData) {
                            rowStream.write(obj);
                        }
                    },
                    success: () => {
                        rowStream.end();
                    },
                    error: (error) => {
                        reject(toError(error));
                    }
                });
            });
        }
        else {
            return new Promise((resolve, reject) => {
                let fullData = [];
                this.client.execute({
                    query,
                    schema: this.config.schema || 'default',
                    data: (error, data, columns) => {
                        const normalData = this.normalizeResultOverColumns(data, columns);
                        fullData = (0, ramda_1.concat)(normalData, fullData);
                    },
                    success: () => {
                        resolve(fullData);
                    },
                    error: (error) => {
                        reject(toError(error));
                    }
                });
            });
        }
    }
    downloadQueryResults(query, values, options) {
        if (options.streamImport) {
            return this.stream(query, values, options);
        }
        return super.downloadQueryResults(query, values, options);
    }
    informationSchemaQuery() {
        if (this.config.schema) {
            return `${super.informationSchemaQuery()} AND columns.table_schema = '${this.config.schema}'`;
        }
        return super.informationSchemaQuery();
    }
    normalizeResultOverColumns(data, columns) {
        const columnNames = (0, ramda_1.map)((0, ramda_1.prop)('name'), columns || []);
        const arrayToObject = (0, ramda_1.zipObj)(columnNames);
        return (0, ramda_1.map)(arrayToObject, data || []);
    }
    stream(query, values, _options) {
        const queryWithParams = this.prepareQueryWithParams(query, values);
        return this.queryPromised(queryWithParams, true);
    }
    capabilities() {
        return {
            unloadWithoutTempTable: true
        };
    }
    async createSchemaIfNotExists(schemaName) {
        await this.query(`CREATE SCHEMA IF NOT EXISTS ${this.config.catalog}.${schemaName}`, []);
    }
    // Export bucket methods
    async isUnloadSupported() {
        return this.config.exportBucket !== undefined;
    }
    async unload(tableName, options) {
        if (!this.config.exportBucket) {
            throw new Error('Export bucket is not configured.');
        }
        if (!SUPPORTED_BUCKET_TYPES.includes(this.config.bucketType)) {
            throw new Error(`Unsupported export bucket type: ${this.config.bucketType}`);
        }
        const types = options.query
            ? await this.unloadWithSql(tableName, options.query.sql, options.query.params)
            : await this.unloadWithTable(tableName);
        const csvFile = await this.getCsvFiles(tableName);
        return {
            exportBucketCsvEscapeSymbol: this.config.exportBucketCsvEscapeSymbol,
            csvFile,
            types,
            csvNoHeader: true,
        };
    }
    splitTableFullName(tableFullName) {
        const [schema, tableName] = tableFullName.split('.');
        return { schema, tableName };
    }
    generateTableColumnsForExport(types) {
        return types.map((c) => `CAST(${c.name} AS varchar) ${c.name}`).join(', ');
    }
    async unloadWithSql(tableFullName, sql, params) {
        return this.unloadGeneric({
            tableFullName,
            typeSql: sql,
            typeParams: params,
            fromSql: sql,
            fromParams: params
        });
    }
    async unloadWithTable(tableFullName) {
        return this.unloadGeneric({
            tableFullName,
            typeSql: `SELECT * FROM ${tableFullName}`,
            typeParams: [],
            fromSql: tableFullName,
            fromParams: []
        });
    }
    async unloadGeneric(params) {
        if (!this.config.exportBucket) {
            throw new Error('Export bucket is not configured.');
        }
        const { bucketType, exportBucket } = this.config;
        const types = await this.queryColumnTypes(params.typeSql, params.typeParams);
        const { schema, tableName } = this.splitTableFullName(params.tableFullName);
        const tableWithCatalogAndSchema = `${this.config.catalog}.${schema}.${tableName}`;
        const protocol = {
            gcs: 'gs',
            s3: this.config.exportBucketS3AdvancedFS ? 's3a' : 's3'
        }[bucketType || 'gcs'];
        const externalLocation = `${protocol}://${exportBucket}/${schema}/${tableName}`;
        const withParams = `( external_location = '${externalLocation}', format = 'CSV')`;
        const select = `SELECT ${this.generateTableColumnsForExport(types)} FROM (${params.fromSql})`;
        const createTableQuery = `CREATE TABLE ${tableWithCatalogAndSchema} WITH ${withParams} AS (${select})`;
        try {
            await this.query(createTableQuery, params.fromParams);
        }
        finally {
            await this.query(`DROP TABLE IF EXISTS ${tableWithCatalogAndSchema}`, []);
        }
        return types;
    }
    async queryColumnTypes(sql, params) {
        const response = await this.stream(`${sql} LIMIT 0`, params || [], { highWaterMark: 1 });
        const result = [];
        for (const column of response.types || []) {
            result.push({ name: column.name, type: this.toGenericType(column.type) });
        }
        return result;
    }
    async getCsvFiles(tableFullName) {
        if (!this.config.exportBucket) {
            throw new Error('Export bucket is not configured.');
        }
        const { bucketType, exportBucket } = this.config;
        const { schema, tableName } = this.splitTableFullName(tableFullName);
        switch (bucketType) {
            case 'gcs':
                return this.extractFilesFromGCS({ credentials: this.config.credentials }, exportBucket, `${schema}/${tableName}`);
            case 's3':
                return this.extractUnloadedFilesFromS3({
                    credentials: this.config.accessKeyId && this.config.secretAccessKey
                        ? {
                            accessKeyId: this.config.accessKeyId,
                            secretAccessKey: this.config.secretAccessKey,
                        }
                        : undefined,
                    region: this.config.exportBucketRegion,
                }, exportBucket, `${schema}/${tableName}`);
            default:
                throw new Error(`Unsupported export bucket type: ${bucketType}`);
        }
    }
}
exports.PrestoDriver = PrestoDriver;
//# sourceMappingURL=PrestoDriver.js.map